/***********************************************************************************

  Program by MrAlvin, October 2010   -   version 1.1   -   mralvin@3iii.dk

  Create Thermistor table for NTC to be used for RepRap, by comparing/managing
  a calibration temperature bed using a K-type AD595 temperatur sensor to
  get raw reading correlation to actual real world temperatures of a NTC.
 
  Temperature readout on LCD and serial connection.
  
  Control of heater by an on/off pin controlling a transistor or MosFET.
  
  ***********************************************************************************
 
  Connect two sensors to a heating bed, one NTC sensor and one K-type sensor 
  (or some other sensor you trust).
  
  Run the heating bed through a temperature range stopping at regular 
  intervals, and stay at each interval for some time, to make sure the
  two temperature sensors are at the exact same temperature.
  
  Take a note of the correlation between raw reading and temperature in celcius
  of the NTC sensor to make a translation table, or copy the resulting 
  table from a terminal program with VT100 emulation, I recommend puTTY, 
  see: http://www.chiark.greenend.org.uk/~sgtatham/putty/download.html

  The Serial Monitor of the Arduino IDE can also be used, but the screen will keep scrolling, 
  whereas puTTY has the "go home" command, so the text will appear stable, with only temperature
  readings changing.
  To use the Serial Monitor, change the #define of PUTTY to false, and set the baud rate in the Serial Monitor to 115200
  
  The numbers at the welcome screen on the LCD are (in order):
  "start temperature" - "stop temperature" : "step size" : "resulting table length"
 
 
  If no start button is used (default) then the calibration proccess starts 
  automatically at boot or reset of the Arduino
 
 
************************************************************************************
 *
 *  Hardware information:
 * 
 *  :: The LCD circuit :: Labitat LCD Backpack style connections
 *   - LCD RS pin4 to digital pin 7
 *   - LCD Enable pin6 to digital pin 6
 *   - LCD D4 pin11 to digital pin 5
 *   - LCD D5 pin12 to digital pin 4
 *   - LCD D6 pin13 to digital pin 3
 *   - LCD D7 pin14 to digital pin 2
 *
 *
 *  :: The temperature sensing circuits ::
 *   - RepRap NTC on analog pin 0
 *   - AD595 K-type temperture sensor on analog pin 1
 *
 *  :: The power managing/heater circuit ::
 *   - On/Off and PWM from HEAT_PIN  (default: digital pin 9)
 *
 *  :: Start button (if used) ::
 *   - connect to low on START_PIN to begin calibration processing (default: digital pin 11) 
 *  
 *************************************************************************/
  
  
    //Temperature range to calibrate
    #define TEMP_START            30  //celcius
    #define TEMP_MAX             290  //celcius
    #define TEMP_STEP             10  //celcius
    #define TEMP_STABLE_DURATION  30  //seconds


// the following defines can be edited, but you may as well just connect your hardware according to 
    #define START_PIN             -1  //use digital numbering - input pin (set to -1 if no start button is used)
    #define HEAT_PIN               9  //use digital numbering - output pin
    #define LED_PIN               13  //use digital numbering - output pin
    
    #define NTC_PIN                0  //use analog numbering  - NTC measure pin
    #define AD_PIN                 1  //use analog numbering  - Thermocouple AD595 measure pin
    
    #define BAUD_RATE         115200  
    #define PUTTY               true  //set to true if you use the terminal program puTTY to listen on the serial port
                                      //set to false if you use the Arduino Serial Monitor to listen on the serial port
                                     
    #define READOUT_SPEED        200  //pause 200ms between each screen update = update screen 5 times a second

  
// include the library code:
    #include <LiquidCrystal.h>
    
// initialize the library with the numbers of the interface pins
   LiquidCrystal lcd(7, 6, 5, 4, 3, 2);     //Labitat.dk  LCD I/O Backpack default pinout
   //LiquidCrystal lcd(12,11, 5, 4, 3, 2);  //Arduino playground default pinout


/************************************************************************
  There should be no need to adjust things below this point              
*/

    
/**** define global variables ***/
  int state = 0;
  unsigned long previous_millis = 0;
  int target_temp = TEMP_START;
  int update_serial = true;
  int i; // used in "for" loops
  boolean heart_beat = true;
  boolean getting_ready = true;
  int raw_avg = 0, raw_min = 1024 , raw_max = 0;

  int reading;
  float voltage;

  //NTC
  int rawNTC;
  float celciusNTC;
    
  // AD595
  int rawAD;
  float celsiusAD;
  
  //Storage variables
  # define MAX_STORAGE  (((TEMP_MAX-TEMP_START)/TEMP_STEP)+2)
  int storage[MAX_STORAGE][2];
  int storage_count = 0;

  

/**********************************************************************  
 **********************************************************************
 **                                                                  **
 **                Arduino main default SETUP and LOOP               **
 **                                                                  **
 **********************************************************************
 **********************************************************************/
 
void setup() {
  // set up the LCD's number of rows and columns: 
  lcd.begin(16, 2);
  
  // Welcome/boot information on LCD 
  lcd.setCursor(0, 0);  // set the cursor to column 0, line 1
  lcd.print("NTC Calibration");
  lcd.setCursor(0, 1);  // set the cursor to column 0, line 2
  lcd.print(TEMP_START);
  lcd.print("-");
  lcd.print(TEMP_MAX);
  lcd.print(":");
  lcd.print(TEMP_STEP);
  lcd.print(":");
  lcd.print(MAX_STORAGE-1);
  //lcd.setCursor(15, 1);  // set the cursor to column 14, line 2
  //lcd.print("5");
  
  //leave welcome information for at least 5 second
  for (int i=5; i>0; i--) {
    lcd.setCursor(15, 1);  // set the cursor to column 0, line 2
    lcd.print(i);
    delay(1000);
  }

  //Initialize output pins
  pinMode(HEAT_PIN, OUTPUT);
  pinMode(LED_PIN, OUTPUT);
  
  //Initialize input pin
  if (START_PIN > -1 ) {
    pinMode(START_PIN, INPUT);
    digitalWrite(START_PIN,HIGH); // turn on internal 20k pull-up resistor
  }
  
  // Initialize serial communiction
  Serial.begin(BAUD_RATE);
  Serial.println("Start");
  
  // Welcome/boot information on Serial port
  if (PUTTY) Serial.println("\x1B[2J" ); //Clear screen in puTTY (terminal program with VT100 emulation)
  Serial.println("NTC Calibration");
  if (PUTTY) for (i=1; i<=7; i++) Serial.println("");
  Serial.print("Temperature calibration range: ");
  Serial.print(TEMP_START);
  Serial.print("-");
  Serial.println(TEMP_MAX);
  Serial.print("In steps: ");
  Serial.println(TEMP_STEP);
  Serial.print("Max Table size: ");
  Serial.print(MAX_STORAGE-1);
  Serial.println("                 ");
  serial_home();
}

void loop() {
  read_temps();
  manage_lcd();
  manage_target_temp();
  manage_heater();
  serial_home();

  //toggle LED_PIN
  if (heart_beat) {digitalWrite(LED_PIN,HIGH); heart_beat=false;} else {digitalWrite(LED_PIN,LOW); heart_beat=true;}
  
  // wait a bit to repeat the loop
  delay(200);
  
}

void serial_home() {
  // update serial terminal/monitor program
  if (update_serial) {
    if (PUTTY) {
    Serial.println("\x1B[H" );  //Home in puTTY (terminal program with VT100 emulation)
    //Serial.println("\x1B[2J" ); //Clear screen in puTTY (terminal program with VT100 emulation)
    } else Serial.println("    ");
  }
}

void manage_target_temp() {
  switch (state) {
    case 0:   // State 0 - wait for start button to be pushed
      if (START_PIN > -1) {
        if (digitalRead(START_PIN) == LOW) state = 1;  //use external button to start and restart calibration process
      } else { //no start_pin defined, so auto-start calibration routine after boot/reset of the Arduino
        if ((celsiusAD <= target_temp) || (rawAD <=1)) state = 1;  // should however waite until probe temperature is below start_temp
      }                                                              // or the control temp analog converter is at its lowest possible point
      break;
    case 1:   // State 1 - wait for reaching target_temp
      if (celsiusAD >= target_temp) {
        if (rawNTC <=1) {
          state = 4; // we are very, very done! This NTC has reached its limit!
        } else {
          state = 2;  //target_temp is reached
          previous_millis = millis();
          getting_ready = false;
        }
      }
      break;
    case 2:   // State 2 - stay at target_temp for TEMP_STABLE_DURATION
      raw_avg = avg(raw_avg, rawNTC); // average rawNTC
      if ((millis() - previous_millis) > (TEMP_STABLE_DURATION * 1000)) state = 3;
      break;
    case 3:   // State 3 - Store found values, and up target_temp with next TEMP_STEP
      storage[storage_count][1]=raw_avg;
      storage[storage_count][2]=target_temp;
      storage_count++;
      raw_avg = 0;
      raw_min = 1024;
      raw_max = 0;
      
      target_temp = target_temp + TEMP_STEP;
      if (target_temp > TEMP_MAX) state = 4; // we are done
      else state = 1; // reach new target temp
      break;
    case 4:   // State 4 - job done 
      target_temp = 0;    // turn off heater
      serial_home();
      if (PUTTY) Serial.println("\x1B[2J" ); //Clear screen in puTTY (terminal program with VT100 emulation)
      update_serial = false;
      print_table();      // print table definition to the serial port
      
      storage_count = 0;
      state = 5;
      break;
    case 5:   //State 5 - leave things on the serial screen, for the user to copy the table
      break;
  }
  //print target temp on serial screen
  if (update_serial) {
    Serial.print("Target Temp: ");
    Serial.print(target_temp);
    if (getting_ready) {
      Serial.println(" - Getting down to temperature start");
    } else Serial.println("                                       ");
    Serial.print("This reading - avg: ");
    Serial.print(raw_avg);
    Serial.print("  min: ");
    Serial.print(raw_min);
    Serial.print("  max: ");
    Serial.print(raw_max);
    Serial.println("        ");
    Serial.println("                                                              ");
    Serial.print("State: ");
    Serial.print(state);
  }   
}

int avg (int a, int b) {
  // set min & max
  if (a < b) {
    if (a > 0) if (a < raw_min) raw_min = a;
    if (b < raw_max) raw_max = b;
  } else {
    if (b < raw_min) raw_min = b;
    if (a > raw_max) raw_max = a;
  }
  
  //set average
  if (a == 0) {
    return b; 
  } else {
    a = a + b;
    a = a / 2;
    return a;
  }
}


/**********************************************************************  
 **********************************************************************
 **                                                                  **
 **                 Read TEMPERATURES procedures                     **
 **                                                                  **
 **********************************************************************
 **********************************************************************/
void read_temps() {

  
   
  /**********************************************************************  
  *   Using the AD595 K-type temperature sensor chip
  ***********************************************************************
      Description and program example from  
      http://www.reprap.org/wiki/Thermocouple_Sensor_1_0
  
  */
  rawAD = analogRead(AD_PIN);
  celsiusAD = ( 5.0 * rawAD * 100.0) / 1024.0;
 
  if (update_serial) {
    Serial.print("AD595     Raw: ");
    Serial.print(rawAD);
    Serial.print(" - Celsius: ");
    Serial.print(celsiusAD,1);
    Serial.println("     ");
  }
  /**********************************************************************  
     Measuring temperature from a NTC resistor on pin 0
  ***********************************************************************/ 

    reading = analogRead(NTC_PIN); 
    rawNTC  =  reading;

    celciusNTC = analog2temp(reading); 
    
    if (update_serial) {
      Serial.print("NTC (0)   Raw: ");
      Serial.print(rawNTC);
      Serial.print(" - Celcius: ");
      Serial.print(celciusNTC,1);
      Serial.println("     ");
    }
  
  
}  
 
    
/**********************************************************************
   LCD Display section
***********************************************************************/
void manage_lcd() {
  
  // set the cursor to column 0, line 1
  // (note: line 1 is the second row, since counting begins with 0):
  
  lcd.setCursor(0, 0);
  lcd.print("AD  R:");
  lcd.print(rawAD);
  lcd.print("   ");
  lcd.setCursor(11, 0);
  lcd.print("C:");
  lcd.print(celsiusAD,0);
  lcd.print("     ");
  
  lcd.setCursor(0, 1);
  lcd.print("NTC R:");
  lcd.print(rawNTC);
  lcd.print("   ");
  
 
  lcd.setCursor(11, 1);
  if (state <= 4) {
    lcd.print("T:");
    lcd.print(target_temp);
    lcd.print("     "); 
  } else {
    lcd.print("Done         ");
  }
  
  
}


void manage_heater() {
  if(celsiusAD >= target_temp) digitalWrite(HEAT_PIN,LOW); //OFF
  else digitalWrite(HEAT_PIN,HIGH); //ON
}



// Takes temperature value as input and returns corresponding analog value from RepRap thermistor temp table.
// This is needed because PID in hydra firmware hovers around a given analog value, not a temp value.
// This function is derived from inversing the logic from a portion of getTemperature() in FiveD RepRap firmware.
float temp2analog(int celsius) {
    int raw = 0;
    byte i;
    
    for (i=1; i<NUMTEMPS; i++) {
      if (temptable[i][1] < celsius) {
        raw = temptable[i-1][0] + 
          (celsius - temptable[i-1][1]) * 
          (temptable[i][0] - temptable[i-1][0]) /
          (temptable[i][1] - temptable[i-1][1]);
      
        break;
      }
    }

    // Overflow: Set to last value in the table
    if (i == NUMTEMPS) raw = temptable[i-1][0];

    return 1023 - raw;
}

// Derived from RepRap FiveD extruder::getTemperature()
float analog2temp(int raw) {
    int celsius = 0;
    byte i;

    for (i=1; i<NUMTEMPS; i++) {
      if (temptable[i][0] > raw)  {
        celsius  = temptable[i-1][1] + 
          (raw - temptable[i-1][0]) * 
          (temptable[i][1] - temptable[i-1][1]) /
          (temptable[i][0] - temptable[i-1][0]);

        break;
      }
    }

    // Overflow: Set to last value in the table
    if (i == NUMTEMPS) celsius = temptable[i-1][1];

    return celsius; 
}

 
void print_table() {
  int i;

  Serial.print("#define NUMTEMPS ");
  Serial.println(storage_count);
  Serial.println("short temptable[NUMTEMPS][2] = {  ");
  
  for (i=storage_count-1; i>0; i--) {
  
    Serial.print("   {");
    Serial.print(storage[i][1]);
    Serial.print(","); 
    Serial.print(storage[i][2]);
    Serial.println("},");
  }
  Serial.print("   {");
  Serial.print(storage[0][1]);
  Serial.print(","); 
  Serial.print(storage[0][2]);
  Serial.println("}");
  Serial.println("};");
}
  
  
/**********************************************************************
 * Thermistor Table
 **********************************************************************
  For thermistor:  www.Rapidonline.com - NTC part no. 104GT-1  - Order code: 61-0452
  More about Thermistor lookup table information for RepRap Temperature Sensor Boards (http://make.rrrf.org/ts) 
*/

#define NUMTEMPS 72
short temptable[NUMTEMPS][2] = {
   {7,355},
   {7,350},
   {8,345},
   {8,340},
   {9,335},
   {10,330},
   {10,325},
   {11,320},
   {12,315},
   {13,310},
   {13,305},
   {14,300},
   {15,295},
   {17,290},
   {18,285},
   {20,280},
   {21,275},
   {23,270},
   {25,265},
   {27,260},
   {29,255},
   {32,250},
   {34,245},
   {37,240},
   {41,235},
   {45,230},
   {49,225},
   {54,220},
   {59,215},
   {64,210},
   {71,205},
   {78,200},
   {85,195},
   {95,190},
   {105,185},
   {116,180},
   {129,175},
   {142,170},
   {158,165},
   {174,160},
   {193,155},
   {213,150},
   {238,145},
   {271,140},
   {295,135},
   {322,130},
   {354,125},
   {385,120},
   {418,115},
   {453,110},
   {490,105},
   {529,100},
   {566,95},
   {605,90},
   {643,85},
   {681,80},
   {717,75},
   {752,70},
   {782,65},
   {814,60},
   {838,55},
   {863,50},
   {883,45},
   {904,40},
   {919,35},
   {934,30},
   {946,25},
   {958,20},
   {966,15},
   {974,10},
   {983,5},
   {992,0}
};

